import aiohttp
import asyncio

from alpaca_trade_api.entity_v2 import BarsV2, QuotesV2, TradesV2, \
    EntityList, TradeV2, QuoteV2
import pandas as pd
from alpaca_trade_api.common import URL, get_credentials, get_data_url


class AsyncRest:
    def __init__(self,
                 key_id: str = None,
                 secret_key: str = None,
                 data_url: URL = None,
                 api_version: str = None,
                 raw_data: bool = False
                 ):
        """
        :param raw_data: should we return api response raw or wrap it with
                         Entity objects.
        """
        self._key_id, self._secret_key, _ = get_credentials(key_id, secret_key)
        self._data_url: URL = URL(data_url or get_data_url())

    def _get_historic_url(self, _type, symbol):
        return f"{self._data_url}/v2/stocks/{symbol}/{_type}"

    def _get_latest_url(self, _type, symbol):
        return f"{self._data_url}/v2/stocks/{symbol}/{_type}/latest"

    async def _iterate_requests(self,
                                symbol,
                                payload,
                                limit,
                                entity_type: str,
                                entity_list_type: EntityList) -> pd.DataFrame:
        """
        iterates the api asynchronously until we get all requested data
        :param symbol:
        :param payload:
        :param entity_type: bars/trades/quotes
        :param entity_list_type:
        :return:
        """
        df = pd.DataFrame({})
        url = self._get_historic_url(entity_type, symbol)
        async for packet in self._request(url, payload):
            if packet.get(entity_type):
                response = entity_list_type(packet[entity_type]).df
                df = pd.concat([df, response], axis=0)
                if len(df) >= limit:
                    break

        return df

    async def get_bars_async(self,
                             symbol,
                             start,
                             end,
                             timeframe,
                             limit=1000,
                             adjustment='raw'):
        _type = "bars"

        payload = {
            "adjustment": adjustment,
            "start":      start,
            "end":        end,
            "timeframe":  timeframe,
            "limit":      limit,
        }
        df = await self._iterate_requests(symbol, payload, limit, _type,
                                          BarsV2)

        return symbol, df

    async def get_trades_async(self, symbol, start, end, limit=1000):
        _type = "trades"

        payload = {
            "start": start,
            "end":   end,
            "limit": limit,
        }
        df = await self._iterate_requests(symbol, payload, limit, _type,
                                          TradesV2)

        return symbol, df

    async def get_quotes_async(self, symbol, start, end, limit=1000):
        _type = "quotes"

        payload = {
            "start": start,
            "end":   end,
            "limit": limit,
        }
        df = await self._iterate_requests(symbol, payload, limit, _type,
                                          QuotesV2)

        return symbol, df

    async def get_latest_trade_async(self, symbol: str) -> TradeV2:
        """
        Get the latest trade for the given symbol
        """
        _type = "trades"
        url = self._get_latest_url(_type, symbol)
        opts = self._get_opts()

        async with aiohttp.ClientSession() as session:
            async with session.get(url, **opts) as response:
                response = await response.json()
                if response.get("quote"):
                    result = TradeV2(response["trade"])
                    return symbol, result

    async def get_latest_quote_async(self, symbol: str) -> QuoteV2:
        """
        Get the latest trade for the given symbol
        """
        _type = "quotes"
        url = self._get_latest_url(_type, symbol)
        opts = self._get_opts()

        async with aiohttp.ClientSession() as session:
            async with session.get(url, **opts) as response:
                response = await response.json()
                if response.get("quote"):
                    result = QuoteV2(response["quote"])
                    return symbol, result

    def _get_opts(self, payload=None):
        headers = {}
        headers['APCA-API-KEY-ID'] = self._key_id
        headers['APCA-API-SECRET-KEY'] = self._secret_key
        opts = {
            'headers':         headers,
            # Since we allow users to set endpoint URL via env var,
            # human error to put non-SSL endpoint could exploit
            # uncanny issues in non-GET request redirecting http->https.
            # It's better to fail early if the URL isn't right.
            'allow_redirects': False,
        }
        opts['params'] = payload

        return opts

    async def _request(self, url, payload):
        opts = self._get_opts(payload)
        async with aiohttp.ClientSession() as session:
            while 1:
                async with session.get(url, **opts) as response:

                    response = await response.json()
                    page_token = response.get('next_page_token')
                    payload["page_token"] = page_token
                    yield response

                    if not page_token:
                        break


async def gather_with_concurrency(n, *tasks):
    """
    when working with python function has limitations on the amount of tasks
    it could handle. for that purpose we use this method that splits the tasks.
    it's a bit slower, but gets the job done.
    Follow the example code to learn how to use that
    """
    semaphore = asyncio.Semaphore(n)

    async def sem_task(task):
        async with semaphore:
            return await task

    return await asyncio.gather(*(sem_task(task) for task in tasks),
                                return_exceptions=True)



import warnings
import alpaca_trade_api as tradeapi
from alpaca_trade_api.rest import APIError

import os
import pytest
import requests_mock


@pytest.fixture(autouse=True)
def delete_base_url_envs():
    if 'APCA_API_BASE_URL' in os.environ:
        del os.environ['APCA_API_BASE_URL']
    if 'APCA_API_DATA_URL' in os.environ:
        del os.environ['APCA_API_DATA_URL']


@pytest.fixture
def reqmock():
    with requests_mock.Mocker() as m:
        yield m


def test_api(reqmock):
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    api_raw = tradeapi.REST('key-id', 'secret-key', api_version='v1',
                            raw_data=True)

    # Get a list of accounts
    reqmock.get('https://api.alpaca.markets/v1/account', text='''
    {
      "id": "904837e3-3b76-47ec-b432-046db621571b",
      "status": "ACTIVE",
      "currency": "USD",
      "cash": "4000.32",
      "cash_withdrawable": "4000.32",
      "portfolio_value": "4321.98",
      "pattern_day_trader": false,
      "trading_blocked": false,
      "transfers_blocked": false,
      "account_blocked": false,
      "created_at": "2018-05-03T06:17:56Z"
    }
''')

    account = api.get_account()
    assert account.status == 'ACTIVE'
    assert 'Account(' in str(account)
    assert type(account) == tradeapi.entity.Account
    assert type(api_raw.get_account()) == dict

    # Get a list of assets
    reqmock.get('https://api.alpaca.markets/v1/assets', text='''[
  {
    "id": "904837e3-3b76-47ec-b432-046db621571b",
    "name": "Apple inc.",
    "asset_class": "us_equity",
    "exchange": "NASDAQ",
    "symbol": "AAPL",
    "status": "active",
    "tradable": true
  }
]''')
    assets = api.list_assets()
    assert assets[0].name == 'Apple inc.'
    assert type(assets[0]) == tradeapi.entity.Asset
    assert type(api_raw.list_assets()) == list

    # Get an asset
    asset_id = '904837e3-3b76-47ec-b432-046db621571b'
    reqmock.get('https://api.alpaca.markets/v1/assets/{}'.format(asset_id),
                text='''{
    "id": "904837e3-3b76-47ec-b432-046db621571b",
    "name": "Apple inc.",
    "asset_class": "us_equity",
    "exchange": "NASDAQ",
    "symbol": "AAPL",
    "status": "active",
    "tradable": true
  }''')
    asset = api.get_asset(asset_id)
    assert asset.name == 'Apple inc.'
    assert type(asset) == tradeapi.entity.Asset
    assert type(api_raw.get_asset(asset_id)) == dict


def test_orders(reqmock):
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    api_raw = tradeapi.REST('key-id', 'secret-key', api_version='v1',
                            raw_data=True)

    # Get a list of orders
    reqmock.get(
        'https://api.alpaca.markets/v1/orders',
        text='''[
  {
    "id": "904837e3-3b76-47ec-b432-046db621571b",
    "client_order_id": "904837e3-3b76-47ec-b432-046db621571b",
    "account_id": "904837e3-3b76-47ec-b432-046db621571b",
    "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
    "qty": "15",
    "notional": null,
    "side": "buy",
    "type": "market",
    "timeinforce": "day",
    "limit_price": "107.00",
    "stop_price": "106.00",
    "filled_avg_price": "107.00",
    "status": "new",
    "created_at": "2018-03-09T19:05:27Z",
    "updated_at": "2018-03-09T19:05:27Z",
    "cancelled_at": "2018-03-09T19:05:27Z",
    "expired_at": "2018-03-09T19:05:27Z",
    "filled_at": "2018-03-09T19:05:27Z",
    "failed_at": "2018-03-09T19:05:27Z",
    "filled_qty": "0",
    "hwm": null,
    "trail_percent": null,
    "trail_price": null,
    "failured_reason": "string",
    "cancel_requested_at": "2018-03-09T19:05:27Z",
    "submitted_at": "2018-03-09T19:05:27Z"
  }
]''')
    orders = api.list_orders('all')
    assert orders[0].type == 'market'
    assert type(orders[0]) == tradeapi.entity.Order
    assert type(api_raw.list_orders()) == list

    # Create an order
    reqmock.post(
        'https://api.alpaca.markets/v1/orders',
        text='''{
  "id": "904837e3-3b76-47ec-b432-046db621571b",
  "client_order_id": "904837e3-3b76-47ec-b432-046db621571b",
  "account_id": "904837e3-3b76-47ec-b432-046db621571b",
  "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
  "qty": "15",
  "notional": null,
  "side": "buy",
  "type": "market",
  "timeinforce": "day",
  "limit_price": "107.00",
  "stop_price": "106.00",
  "filled_avg_price": "107.00",
  "status": "new",
  "created_at": "2018-03-09T19:05:27Z",
  "updated_at": "2018-03-09T19:05:27Z",
  "cancelled_at": "2018-03-09T19:05:27Z",
  "expired_at": "2018-03-09T19:05:27Z",
  "filled_at": "2018-03-09T19:05:27Z",
  "failed_at": "2018-03-09T19:05:27Z",
  "filled_qty": "0",
  "hwm": null,
  "trail_percent": null,
  "trail_price": null,
  "failured_reason": "string",
  "cancel_requested_at": "2018-03-09T19:05:27Z",
  "submitted_at": "2018-03-09T19:05:27Z"
}''')
    order = api.submit_order(
        symbol='904837e3-3b76-47ec-b432-046db621571b',
        qty=15,
        side='buy',
        type='market',
        time_in_force='day',
        limit_price='107.00',
        stop_price='106.00',
        client_order_id='904837e3-3b76-47ec-b432-046db621571b',
    )
    assert order.qty == "15"
    assert order.notional is None
    assert order.created_at.hour == 19
    assert type(order) == tradeapi.entity.Order
    assert type(api_raw.submit_order(
        symbol='904837e3-3b76-47ec-b432-046db621571b',
        qty=15,
        side='buy',
        type='market',
        time_in_force='day',
        limit_price='107.00',
        stop_price='106.00',
        client_order_id='904837e3-3b76-47ec-b432-046db621571b',)) == dict
    # now let's test some different acceptable "float" formats
    api.submit_order(
        symbol='904837e3-3b76-47ec-b432-046db621571b',
        qty=15,
        side='buy',
        type='market',
        time_in_force='day',
        limit_price='107.00',  # str float
        stop_price=106.00,  # float
        client_order_id='904837e3-3b76-47ec-b432-046db621571b',
    )
    api.submit_order(
        symbol='904837e3-3b76-47ec-b432-046db621571b',
        qty=15,
        side='buy',
        type='market',
        time_in_force='day',
        limit_price='107.00',  # str float
        stop_price=106,  # int
        client_order_id='904837e3-3b76-47ec-b432-046db621571b',
    )
    with pytest.raises(ValueError):
        api.submit_order(
            symbol='904837e3-3b76-47ec-b432-046db621571b',
            qty=15,
            side='buy',
            type='market',
            time_in_force='day',
            limit_price='1',
            stop_price="a",
            client_order_id='904837e3-3b76-47ec-b432-046db621571b',
        )
    with pytest.raises(ValueError):
        api.submit_order(
            symbol='904837e3-3b76-47ec-b432-046db621571b',
            qty=15,
            side='buy',
            type='market',
            time_in_force='day',
            limit_price='a',
            stop_price=106.00,
            client_order_id='904837e3-3b76-47ec-b432-046db621571b',
        )

    # Get an order by client order id
    client_order_id = 'client-order-id'
    reqmock.get(
        'https://api.alpaca.markets/v1/orders:by_client_order_id?'
        'client_order_id={}'.format(
            client_order_id, ), text='''{
  "id": "904837e3-3b76-47ec-b432-046db621571b",
  "client_order_id": "904837e3-3b76-47ec-b432-046db621571b",
  "account_id": "904837e3-3b76-47ec-b432-046db621571b",
  "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
  "qty": "15",
  "notional": null,
  "side": "buy",
  "type": "market",
  "timeinforce": "day",
  "limit_price": "107.00",
  "stop_price": "106.00",
  "filled_avg_price": "107.00",
  "status": "new",
  "created_at": "2018-03-09T05:50:50Z",
  "updated_at": "2018-03-09T05:50:50Z",
  "cancelled_at": "2018-03-09T05:50:50Z",
  "expired_at": "2018-03-09T05:50:50Z",
  "filled_at": "2018-03-09T05:50:50Z",
  "failed_at": "2018-03-09T05:50:50Z",
  "filled_qty": "0",
  "hwm": null,
  "trail_percent": null,
  "trail_price": null,
  "failured_reason": "string",
  "cancel_requested_at": "2018-03-09T05:50:50Z",
  "submitted_at": "2018-03-09T05:50:50Z"
}''')
    order = api.get_order_by_client_order_id(client_order_id)
    assert order.submitted_at.minute == 50
    assert type(order) == tradeapi.entity.Order
    assert type(api_raw.get_order_by_client_order_id(client_order_id)) == dict

    # Get an order
    order_id = '904837e3-3b76-47ec-b432-046db621571b'
    reqmock.get(
        'https://api.alpaca.markets/v1/orders/{}'.format(order_id),
        text='''{
  "id": "904837e3-3b76-47ec-b432-046db621571b",
  "client_order_id": "904837e3-3b76-47ec-b432-046db621571b",
  "account_id": "904837e3-3b76-47ec-b432-046db621571b",
  "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
  "qty": 15,
  "notional": null,
  "side": "buy",
  "type": "market",
  "timeinforce": "day",
  "limit_price": "107.00",
  "stop_price": "106.00",
  "filled_avg_price": "107.00",
  "status": "new",
  "created_at": "2018-03-09T05:50:50Z",
  "updated_at": "2018-03-09T05:50:50Z",
  "cancelled_at": "2018-03-09T05:50:50Z",
  "expired_at": "2018-03-09T05:50:50Z",
  "filled_at": "2018-03-09T05:50:50Z",
  "failed_at": "2018-03-09T05:50:50Z",
  "filled_qty": "0",
  "hwm": null,
  "trail_percent": null,
  "trail_price": null,
  "failured_reason": "string",
  "cancel_requested_at": "2018-03-09T05:50:50Z",
  "submitted_at": "2018-03-09T05:50:50Z"
}'''
    )
    order = api.get_order(order_id)
    assert order.side == 'buy'
    assert type(order) == tradeapi.entity.Order
    assert type(api_raw.get_order(order_id)) == dict

    # Cancel an order
    order_id = '904837e3-3b76-47ec-b432-046db621571b'
    reqmock.delete(
        'https://api.alpaca.markets/v1/orders/{}'.format(order_id),
        text='',
        status_code=204,
    )
    api.cancel_order(order_id)


def test_positions(reqmock):
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    api_raw = tradeapi.REST('key-id', 'secret-key', api_version='v1',
                            raw_data=True)

    # Get a list of positions
    reqmock.get(
        'https://api.alpaca.markets/v1/positions',
        text='''[
  {
    "account_id": "904837e3-3b76-47ec-b432-046db621571b",
    "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
    "entry_price": "100.0",
    "qty": "5",
    "side": "long",
    "market_value": "600.0",
    "cost_basis": "500.0",
    "last_price": "120.00"
  }
]'''
    )
    positions = api.list_positions()
    assert positions[0].entry_price == '100.0'
    assert type(positions[0]) == tradeapi.entity.Position
    assert type(api_raw.list_positions()) == list

    # Get an open position
    asset_id = 'test-asset'
    reqmock.get(
        'https://api.alpaca.markets/v1/positions/{}'.format(asset_id),
        text='''{
  "account_id": "904837e3-3b76-47ec-b432-046db621571b",
  "asset_id": "904837e3-3b76-47ec-b432-046db621571b",
  "entry_price": "100.0",
  "qty": "5",
  "side": "long",
  "market_value": "600.0",
  "cost_basis": "500.0",
  "last_price": "120.00"
}'''
    )
    position = api.get_position(asset_id)
    assert position.cost_basis == '500.0'
    assert type(position) == tradeapi.entity.Position
    assert type(api_raw.get_position(asset_id)) == dict


def test_chronos(reqmock):
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    api_raw = tradeapi.REST('key-id', 'secret-key', api_version='v1',
                            raw_data=True)

    # clock
    reqmock.get(
        'https://api.alpaca.markets/v1/clock',
        text='''{
  "timestamp": "2018-04-01T12:00:00.000Z",
  "is_open": true,
  "next_open": "2018-04-01T12:00:00.000Z",
  "next_close": "2018-04-01T12:00:00.000Z"
}'''
    )
    clock = api.get_clock()
    assert clock.is_open
    assert clock.next_open.day == 1
    assert type(clock) == tradeapi.entity.Clock
    assert type(api_raw.get_clock()) == dict

    # calendar
    reqmock.get(
        'https://api.alpaca.markets/v1/calendar?start=2018-01-03',
        text='''[
  {
    "date": "2018-01-03",
    "open": "09:30",
    "close": "16:00"
  }
]'''
    )
    calendar = api.get_calendar(start='2018-01-03')
    assert calendar[0].date.day == 3
    assert calendar[0].open.minute == 30
    assert type(calendar) == list
    assert type(calendar[0]) == tradeapi.entity.Calendar
    assert type(api_raw.get_calendar(start='2018-01-03')) == list


def test_data(reqmock):
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    api_raw = tradeapi.REST('key-id', 'secret-key', api_version='v1',
                            raw_data=True)

    # Latest trade
    reqmock.get(
        'https://data.alpaca.markets/v2/stocks/AAPL/trades/latest',
        text='''
        {
            "symbol": "AAPL",
            "trade": {
                "t": "2021-04-20T12:40:34.123456789Z",
                "x": "J",
                "p": 134.7,
                "s": 20,
                "c": [
                    "@",
                    "T",
                    "I"
                ],
                "i": 32,
                "z": "C"
            }
        }
        '''
    )
    latest_trade = api.get_latest_trade('AAPL')
    assert latest_trade.exchange == "J"
    assert latest_trade.price == 134.7
    assert latest_trade.size == 20
    assert latest_trade.conditions == ["@", "T", "I"]
    assert latest_trade.id == 32
    assert latest_trade.tape == "C"
    assert latest_trade.timestamp.day == 20
    assert latest_trade.timestamp.second == 34
    assert latest_trade.timestamp.nanosecond == 789
    assert type(latest_trade) == tradeapi.entity_v2.TradeV2
    assert type(api_raw.get_latest_trade('AAPL')) == dict

    # Latest quote
    reqmock.get(
        'https://data.alpaca.markets/v2/stocks/AAPL/quotes/latest',
        text='''
        {
            "symbol": "AAPL",
            "quote": {
                "t": "2021-04-20T13:01:57.123456789",
                "ax": "Q",
                "ap": 134.68,
                "as": 1,
                "bx": "K",
                "bp": 134.66,
                "bs": 29,
                "c": [
                    "R"
                ],
                "z": "C"
            }
        }'''
    )

    latest_quote = api.get_latest_quote('AAPL')
    assert latest_quote.ask_exchange == "Q"
    assert latest_quote.ask_price == 134.68
    assert latest_quote.ask_size == 1
    assert latest_quote.bid_exchange == "K"
    assert latest_quote.bid_price == 134.66
    assert latest_quote.bid_size == 29
    assert latest_quote.conditions == ["R"]
    assert latest_quote.tape == "C"
    assert latest_quote.timestamp.day == 20
    assert latest_quote.timestamp.nanosecond == 789
    assert type(latest_quote) == tradeapi.entity_v2.QuoteV2
    assert type(api_raw.get_latest_quote('AAPL')) == dict

    # Snapshot
    reqmock.get(
        'https://data.alpaca.markets/v2/stocks/AAPL/snapshot',
        text='''
        {
            "symbol": "AAPL",
            "latestTrade": {
                "t": "2021-05-03T14:45:50.456Z",
                "x": "D",
                "p": 133.55,
                "s": 200,
                "c": [
                    "@"
                ],
                "i": 61462,
                "z": "C"
            },
            "latestQuote": {
                "t": "2021-05-03T14:45:50.532316972Z",
                "ax": "P",
                "ap": 133.55,
                "as": 7,
                "bx": "Q",
                "bp": 133.54,
                "bs": 9,
                "c": [
                    "R"
                ]
            },
            "minuteBar": {
                "t": "2021-05-03T14:44:00Z",
                "o": 133.485,
                "h": 133.4939,
                "l": 133.42,
                "c": 133.445,
                "v": 182818
            },
            "dailyBar": {
                "t": "2021-05-03T04:00:00Z",
                "o": 132.04,
                "h": 134.07,
                "l": 131.83,
                "c": 133.445,
                "v": 25094213
            },
            "prevDailyBar": {
                "t": "2021-04-30T04:00:00Z",
                "o": 131.82,
                "h": 133.56,
                "l": 131.065,
                "c": 131.46,
                "v": 109506363
            }
        }'''
    )
    snapshot = api.get_snapshot('AAPL')
    assert snapshot.latest_trade.price == 133.55
    assert snapshot.latest_quote.bid_size == 9
    assert snapshot.minute_bar.open == 133.485
    assert snapshot.daily_bar.high == 134.07
    assert snapshot.prev_daily_bar.volume == 109506363

    # Snapshots
    reqmock.get(
        'https://data.alpaca.markets/v2/stocks/snapshots' +
        '?symbols=AAPL,MSFT,INVALID',
        text='''
        {
            "AAPL": {
                "latestTrade": {
                    "t": "2021-05-03T14:48:06.563Z",
                    "x": "D",
                    "p": 133.4201,
                    "s": 145,
                    "c": [
                        "@"
                    ],
                    "i": 62700,
                    "z": "C"
                },
                "latestQuote": {
                    "t": "2021-05-03T14:48:07.257820915Z",
                    "ax": "Q",
                    "ap": 133.43,
                    "as": 7,
                    "bx": "Q",
                    "bp": 133.42,
                    "bs": 15,
                    "c": [
                        "R"
                    ]
                },
                "minuteBar": {
                    "t": "2021-05-03T14:47:00Z",
                    "o": 133.4401,
                    "h": 133.48,
                    "l": 133.37,
                    "c": 133.42,
                    "v": 207020
                },
                "dailyBar": {
                    "t": "2021-05-03T04:00:00Z",
                    "o": 132.04,
                    "h": 134.07,
                    "l": 131.83,
                    "c": 133.42,
                    "v": 25846800
                },
                "prevDailyBar": {
                    "t": "2021-04-30T04:00:00Z",
                    "o": 131.82,
                    "h": 133.56,
                    "l": 131.065,
                    "c": 131.46,
                    "v": 109506363
                }
            },
            "MSFT": {
                "latestTrade": {
                    "t": "2021-05-03T14:48:06.36Z",
                    "x": "D",
                    "p": 253.8738,
                    "s": 100,
                    "c": [
                        "@"
                    ],
                    "i": 22973,
                    "z": "C"
                },
                "latestQuote": {
                    "t": "2021-05-03T14:48:07.243353456Z",
                    "ax": "N",
                    "ap": 253.89,
                    "as": 2,
                    "bx": "Q",
                    "bp": 253.87,
                    "bs": 2,
                    "c": [
                        "R"
                    ]
                },
                "minuteBar": {
                    "t": "2021-05-03T14:47:00Z",
                    "o": 253.78,
                    "h": 253.869,
                    "l": 253.78,
                    "c": 253.855,
                    "v": 25717
                },
                "dailyBar": {
                    "t": "2021-05-03T04:00:00Z",
                    "o": 253.34,
                    "h": 254.35,
                    "l": 251.8,
                    "c": 253.855,
                    "v": 6100459
                },
                "prevDailyBar": null
            },
            "INVALID": null
        }
        '''
    )
    snapshots = api.get_snapshots(['AAPL', 'MSFT', 'INVALID'])
    assert len(snapshots) == 3
    aapl_snapshot = snapshots.get('AAPL')
    assert aapl_snapshot is not None
    assert aapl_snapshot.latest_trade.size == 145
    assert aapl_snapshot.latest_quote.bid_exchange == "Q"
    msft_snapshot = snapshots.get('MSFT')
    assert msft_snapshot is not None
    assert msft_snapshot.minute_bar.low == 253.78
    assert msft_snapshot.daily_bar.close == 253.855
    assert msft_snapshot.prev_daily_bar is None
    assert snapshots.get('INVALID') is None

    # News
    reqmock.get(
        'https://data.alpaca.markets/v1beta1/news' +
        '?symbols=AAPL,TSLA&limit=2',
        text='''
        {
            "news": [
                {
                    "id": 24994117,
                    "headline": "'Tesla Approved...",
                    "author": "Benzinga Newsdesk",
                    "created_at": "2022-01-11T13:50:47Z",
                    "updated_at": "2022-01-11T13:50:47Z",
                    "summary": "",
                    "url": "https://www.benzinga.com/news/some/path",
                    "images": [],
                    "symbols": [
                        "TSLA"
                    ],
                    "source": "benzinga"
                },
                {
                    "id": 24993189,
                    "headline": "Dogecoin Is Down 80% ...",
                    "author": "Samyuktha Sriram",
                    "created_at": "2022-01-11T13:49:40Z",
                    "updated_at": "2022-01-11T13:49:41Z",
                    "summary": "Popular meme-based cryptocurrency...",
                    "url": "https://www.benzinga.com/markets/some/path",
                    "images": [
                        {
                            "size": "large",
                            "url": "https://cdn.benzinga.com/files/some.jpeg"
                        },
                        {
                            "size": "small",
                            "url": "https://cdn.benzinga.com/files/some.jpeg"
                        },
                        {
                            "size": "thumb",
                            "url": "https://cdn.benzinga.com/files/some.jpeg"
                        }
                    ],
                    "symbols": [
                        "BTCUSD",
                        "DOGEUSD",
                        "SHIBUSD",
                        "TSLA"
                    ],
                    "source": "benzinga"
                }
            ]
        }
        '''
    )
    news = api.get_news(['AAPL', 'TSLA'], limit=2)
    assert len(news) == 2
    first = news[0]
    assert first is not None
    assert first.author == 'Benzinga Newsdesk'
    assert 'TSLA' in first.symbols
    assert first.source == 'benzinga'
    assert type(first) == tradeapi.entity_v2.NewsV2
    second = news[1]
    assert second is not None
    assert second.headline != ''
    assert type(second.images) == list
    assert 'TSLA' in second.symbols
    assert second.source == 'benzinga'
    assert type(second) == tradeapi.entity_v2.NewsV2


def test_timeframe(reqmock):
    # Custom timeframe: Minutes
    reqmock.get('https://data.alpaca.markets/v2/stocks/AAPL/bars?'
                'timeframe=45Min&adjustment=raw&'
                'start=2021-06-08&end=2021-06-08', text='{}')
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    timeframe = tradeapi.TimeFrame(45, tradeapi.TimeFrameUnit.Minute)
    api.get_bars('AAPL', timeframe, '2021-06-08', '2021-06-08')
    assert reqmock.called

    # Custom timeframe: Hours
    reqmock.get('https://data.alpaca.markets/v2/stocks/AAPL/bars?'
                'timeframe=23Hour&adjustment=raw&'
                'start=2021-06-08&end=2021-06-08', text='{}')
    timeframe = tradeapi.TimeFrame(23, tradeapi.TimeFrameUnit.Hour)
    api.get_bars('AAPL', timeframe, '2021-06-08', '2021-06-08')
    assert reqmock.called

    # Cannot be initialized at invalid combinations
    with pytest.raises(Exception):
        tradeapi.TimeFrame(30, tradeapi.TimeFrameUnit.Hour)

    # Cannot be initialized at invalid combinations
    with pytest.raises(Exception):
        tradeapi.TimeFrame(-1, tradeapi.TimeFrameUnit.Minute)

    # Can be modified after set
    timeframe = tradeapi.TimeFrame(23, tradeapi.TimeFrameUnit.Hour)
    timeframe.amount = 5
    timeframe.unit = tradeapi.TimeFrameUnit.Minute
    reqmock.get('https://data.alpaca.markets/v2/stocks/AAPL/bars?'
                'timeframe=5Min&adjustment=raw&'
                'start=2021-06-08&end=2021-06-08', text='{}')
    api.get_bars('AAPL', timeframe, '2021-06-08', '2021-06-08')
    assert reqmock.called

    # Cannot be modified to an invalid range
    timeframe = tradeapi.TimeFrame(23, tradeapi.TimeFrameUnit.Hour)
    with pytest.raises(Exception):
        timeframe.amount = 30

    # Cannot be modified to an invalid range
    timeframe = tradeapi.TimeFrame(23, tradeapi.TimeFrameUnit.Hour)
    timeframe.unit = tradeapi.TimeFrameUnit.Minute
    timeframe.amount = 59
    with pytest.raises(Exception):
        timeframe.unit = tradeapi.TimeFrameUnit.Hour


def test_watchlists(reqmock):
    api = tradeapi.REST('key-id', 'secret-key', api_version='v1')
    api_raw = tradeapi.REST('key-id', 'secret-key', api_version='v1',
                            raw_data=True)
    # get watchlists
    reqmock.get(
        'https://api.alpaca.markets/v1/watchlists',
        text='''[
    {
        "id": "900e20b1-46eb-492b-a505-2ea67386b5fd",
        "account_id": "1f893862-13b5-4603-b3ca-513980c00c6e",
        "created_at": "2019-10-31T01:45:41.308091Z",
        "updated_at": "2019-12-09T17:50:57.151693Z",
        "name": "Primary Watchlist"
    },
    {
        "id": "e65f2f2d-b596-4db6-bd68-1b7ceb77cccc",
        "account_id": "1f893862-13b5-4603-b3ca-513980c00c6e",
        "created_at": "2020-01-23T00:52:07.049138Z",
        "updated_at": "2020-01-23T00:57:27.063889Z",
        "name": "dev"
    },
    {
        "id": "e7574813-a853-4536-a52b-b47cc25def14",
        "account_id": "1f893862-13b5-4603-b3ca-513980c00c6e",
        "created_at": "2020-01-23T01:36:25.807997Z",
        "updated_at": "2020-01-23T01:36:25.807997Z",
        "name": "prod"
    }
    ]''')
    watchlists = api.get_watchlists()
    assert watchlists[0].name == 'Primary Watchlist'
    assert watchlists[1].id == 'e65f2f2d-b596-4db6-bd68-1b7ceb77cccc'
    assert watchlists[2].account_id == '1f893862-13b5-4603-b3ca-513980c00c6e'
    assert type(watchlists[0]) == tradeapi.entity.Watchlist
    assert type(api_raw.get_watchlists()) == list

    # get a watchlist by watchlist_id
    watchlist_id = "e65f2f2d-b596-4db6-bd68-1b7ceb77cccc"
    symbol = "AMD"
    reqmock.get(
        'https://api.alpaca.markets/v1/watchlists/{}'.format(watchlist_id),
        text='''{
            "id": "e65f2f2d-b596-4db6-bd68-1b7ceb77cccc",
            "account_id": "1f893862-13b5-4603-b3ca-513980c00c6e",
            "created_at": "2020-01-23T00:52:07.049138Z",
            "updated_at": "2020-01-23T00:57:27.063889Z",
            "name": "Primary Watchlist",
            "assets": [
        {
            "id": "03fb07bb-5db1-4077-8dea-5d711b272625",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "AMD",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        },
        {
            "id": "4ce9353c-66d1-46c2-898f-fce867ab0247",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "NVDA",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        },
        {
            "id": "bb2a26c0-4c77-4801-8afc-82e8142ac7b8",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "NFLX",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        },
        {
            "id": "24cbba8c-831b-44e2-8503-dd0c2ed7af8f",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "PYPL",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        }
    ]
    }''')
    watchlist = api.get_watchlist(watchlist_id)
    assert watchlist.name == 'Primary Watchlist'
    assert len(watchlist.assets) == 4
    assert watchlist.assets[0]["id"] == "03fb07bb-5db1-4077-8dea-5d711b272625"
    assert watchlist.assets[0]["class"] == "us_equity"
    assert watchlist.assets[0]["exchange"] == "NASDAQ"
    assert watchlist.assets[0]["symbol"] == "AMD"
    assert watchlist.assets[0]["name"] == ""
    assert watchlist.assets[0]["status"] == "active"
    assert watchlist.assets[0]["tradable"]
    assert watchlist.assets[0]["marginable"]
    assert watchlist.assets[0]["shortable"]
    assert watchlist.assets[0]["easy_to_borrow"]
    assert type(watchlist) == tradeapi.entity.Watchlist
    assert type(api_raw.get_watchlist(watchlist_id)) == dict

    # add an asset to a watchlist
    reqmock.post(
        'https://api.alpaca.markets/v1/watchlists/{}'.format(watchlist_id),
        text='''{
            "id": "e65f2f2d-b596-4db6-bd68-1b7ceb77cccc",
            "account_id": "1f893862-13b5-4603-b3ca-513980c00c6e",
            "created_at": "2020-01-23T00:52:07.049138Z",
            "updated_at": "2020-01-23T00:57:27.063889Z",
            "name": "Primary Watchlist",
            "assets": [
        {
            "id": "03fb07bb-5db1-4077-8dea-5d711b272625",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "AMD",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        }
    ]
    }''')
    watchlist = api.add_to_watchlist(watchlist_id, symbol="AMD")
    assert watchlist.name == 'Primary Watchlist'
    assert len(watchlist.assets) == 1
    assert watchlist.assets[0]["symbol"] == "AMD"
    assert type(watchlist) == tradeapi.entity.Watchlist
    assert type(api_raw.add_to_watchlist(watchlist_id, symbol="AMD")) == dict

    # remove an item from a watchlist
    reqmock.delete(
        'https://api.alpaca.markets/v1/watchlists/{}/{}'.format(
            watchlist_id, symbol
        ),
        text='''{
    "id": "e65f2f2d-b596-4db6-bd68-1b7ceb77cccc",
    "account_id": "1f893862-13b5-4603-b3ca-513980c00c6e",
    "created_at": "2020-01-23T00:52:07.049138Z",
    "updated_at": "2020-01-23T00:57:27.063889Z",
    "name": "dev",
    "assets": [
        {
            "id": "03fb07bb-5db1-4077-8dea-5d711b272625",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "AMD",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        },
        {
            "id": "4ce9353c-66d1-46c2-898f-fce867ab0247",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "NVDA",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        },
        {
            "id": "bb2a26c0-4c77-4801-8afc-82e8142ac7b8",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "NFLX",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        },
        {
            "id": "24cbba8c-831b-44e2-8503-dd0c2ed7af8f",
            "class": "us_equity",
            "exchange": "NASDAQ",
            "symbol": "PYPL",
            "name": "",
            "status": "active",
            "tradable": true,
            "marginable": true,
            "shortable": true,
            "easy_to_borrow": true
        }
    ]
    }''')
    api.delete_from_watchlist(watchlist_id, symbol)

    # delete a watchlist
    reqmock.delete(
        'https://api.alpaca.markets/v1/watchlists/{}'.format(watchlist_id),
        text='',
        status_code=204,
    )
    api.delete_watchlist(watchlist_id)


def test_errors(reqmock):
    api_v1 = tradeapi.REST('key-id', 'secret-key', api_version='v1')

    api_v1._retry = 1
    api_v1._retry_wait = 0

    api_v1._do_error = True

    def callback_429(request, context):
        if api_v1._do_error:
            api_v1._do_error = False
            context.status_code = 429
            return 'Too Many Requests'
        else:
            context.status_code = 200
            return '''
    {
      "id": "904837e3-3b76-47ec-b432-046db621571b",
      "status": "ACTIVE",
      "currency": "USD",
      "cash": "4000.32",
      "cash_withdrawable": "4000.32",
      "portfolio_value": "4321.98",
      "pattern_day_trader": false,
      "trading_blocked": false,
      "transfers_blocked": false,
      "account_blocked": false,
      "created_at": "2018-05-03T06:17:56Z"
    }
'''
    # Too Many Requests
    reqmock.get(
        'https://api.alpaca.markets/v1/account',
        text=callback_429,
    )

    account = api_v1.get_account()
    assert account.cash == '4000.32'

    # General API Error
    reqmock.post(
        'https://api.alpaca.markets/v1/orders',
        status_code=403,
        text='''
    {"code": 10041, "message": "Order failed"}
'''
    )

    try:
        api_v1.submit_order(
            symbol='AAPL',
            side='buy',
            qty='3',
            type='market',
            time_in_force='day',
        )
    except APIError as err:
        assert err.code == 10041
        assert err.status_code == 403
        assert err.request is not None
        assert err.response.status_code == err.status_code
    else:
        assert False

    api_v2 = tradeapi.REST('key-id', 'secret-key', api_version='v2')

    # `qty` and `notional` both null in submit_order
    reqmock.post(
        'https://api.alpaca.markets/v2/orders',
        status_code=422,
        text='''
    {"code":40010001,"message":"qty or notional is required"}
'''
    )

    try:
        api_v2.submit_order(
            symbol='AAPL',
            side='buy',
            type='market',
            time_in_force='day',
        )
    except APIError as err:
        assert err.code == 40010001
        assert err.status_code == 422
        assert err.request is not None
        assert err.response.status_code == err.status_code
    else:
        assert False

    # `qty` and `notional` both non-null in submit_order
    reqmock.post(
        'https://api.alpaca.markets/v2/orders',
        status_code=422,
        text='''
    {"code": 40010001, "message": "only one of qty or notional is accepted"}
'''
    )

    try:
        api_v2.submit_order(
            symbol='AAPL',
            side='buy',
            type='market',
            time_in_force='day',
            qty=1,
            notional=1,
        )
    except APIError as err:
        assert err.code == 40010001
        assert err.status_code == 422
        assert err.request is not None
        assert err.response.status_code == err.status_code
    else:
        assert False

    # fractional `qty` passed to replace_order
    reqmock.post(
        'https://api.alpaca.markets/v2/orders',
        text='''{
    "id": "fb61d316-2179-4df2-8b28-eb026c0dd78e",
    "client_order_id": "6de7d1b2-f772-4a0d-8c15-bacea15eb29e",
    "created_at": "2021-04-07T18:25:30.812371Z",
    "updated_at": "2021-04-07T18:25:30.812371Z",
    "submitted_at": "2021-04-07T18:25:30.803178Z",
    "filled_at": null,
    "expired_at": null,
    "canceled_at": null,
    "failed_at": null,
    "replaced_at": null,
    "replaced_by": null,
    "replaces": null,
    "asset_id": "b28f4066-5c6d-479b-a2af-85dc1a8f16fb",
    "symbol": "SPY",
    "asset_class": "us_equity",
    "notional": null,
    "qty": "1",
    "filled_qty": "0",
    "filled_avg_price": null,
    "order_class": "",
    "order_type": "limit",
    "type": "limit",
    "side": "buy",
    "time_in_force": "day",
    "limit_price": "400",
    "stop_price": null,
    "status": "accepted",
    "extended_hours": false,
    "legs": null,
    "trail_percent": null,
    "trail_price": null,
    "hwm": null
}''')
    order = api_v2.submit_order(
        symbol='SPY',
        qty=1,
        side='buy',
        type='limit',
        time_in_force='day',
        limit_price='400.00',
    )

    reqmock.patch(
        'https://api.alpaca.markets/v2/orders/{}'.format(order.id),
        status_code=422,
        text='''{
    "code": 40010001,
    "message": "qty must be integer"
}''')
    try:
        api_v2.replace_order(
            order_id=order.id,
            qty="1.5",
            client_order_id=order.client_order_id,
        )
    except APIError as err:
        assert err.code == 40010001
        assert err.status_code == 422
        assert err.request is not None
        assert err.response.status_code == err.status_code


def test_no_resource_warning_with_context_manager():
    with warnings.catch_warnings():  # ensure no warnings are raised
        warnings.simplefilter("error")
        with tradeapi.REST("key-id", "secret-key", api_version="v1") as api:
            assert api



import asyncio
from collections import defaultdict
import logging
import json
from typing import Dict, List, Optional
import msgpack
import re
import websockets
import queue

from .common import get_base_url, get_data_stream_url, get_credentials, URL
from .entity import Entity
from .entity_v2 import (
    quote_mapping_v2,
    trade_mapping_v2,
    bar_mapping_v2,
    status_mapping_v2,
    luld_mapping_v2,
    cancel_error_mapping_v2,
    correction_mapping_v2,
    orderbook_mapping_v2,
    Trade,
    Quote,
    Bar,
    StatusV2,
    LULDV2,
    CancelErrorV2,
    CorrectionV2,
    NewsV2,
    OrderbookV2,
)

log = logging.getLogger(__name__)

# Default Params we pass to the websocket constructors
WEBSOCKET_DEFAULTS = {
    "ping_interval": 10,
    "ping_timeout": 180,
    "max_queue": 1024,
}


def _ensure_coroutine(handler):
    if not asyncio.iscoroutinefunction(handler):
        raise ValueError('handler must be a coroutine function')


class _DataStream:
    def __init__(self,
                 endpoint: str,
                 key_id: str,
                 secret_key: str,
                 raw_data: bool = False,
                 websocket_params: Optional[Dict] = None) -> None:
        self._endpoint = endpoint
        self._key_id = key_id
        self._secret_key = secret_key
        self._ws = None
        self._running = False
        self._loop = None
        self._raw_data = raw_data
        self._stop_stream_queue = queue.Queue()
        self._handlers = {
            'trades':      {},
            'quotes':      {},
            'bars':        {},
            'updatedBars': {},
            'dailyBars':   {},
        }
        self._name = 'data'
        self._should_run = True
        self._max_frame_size = 32768
        self._websocket_params = websocket_params

        if self._websocket_params is None:
            self._websocket_params = WEBSOCKET_DEFAULTS

    async def _connect(self):
        self._ws = await websockets.connect(
            self._endpoint,
            extra_headers={'Content-Type': 'application/msgpack'},
            **self._websocket_params
        )
        r = await self._ws.recv()
        msg = msgpack.unpackb(r)
        if msg[0]['T'] != 'success' or msg[0]['msg'] != 'connected':
            raise ValueError('connected message not received')

    async def _auth(self):
        await self._ws.send(
            msgpack.packb({
                'action': 'auth',
                'key':    self._key_id,
                'secret': self._secret_key,
            }))
        r = await self._ws.recv()
        msg = msgpack.unpackb(r)
        if msg[0]['T'] == 'error':
            raise ValueError(msg[0].get('msg', 'auth failed'))
        if msg[0]['T'] != 'success' or msg[0]['msg'] != 'authenticated':
            raise ValueError('failed to authenticate')

    async def _start_ws(self):
        await self._connect()
        await self._auth()
        log.info(f'connected to: {self._endpoint}')

    async def close(self):
        if self._ws:
            await self._ws.close()
            self._ws = None
            self._running = False

    async def stop_ws(self):
        self._should_run = False
        if self._stop_stream_queue.empty():
            self._stop_stream_queue.put_nowait({"should_stop": True})

    async def _consume(self):
        while True:
            if not self._stop_stream_queue.empty():
                self._stop_stream_queue.get(timeout=1)
                await self.close()
                break
            else:
                try:
                    r = await asyncio.wait_for(self._ws.recv(), 5)
                    msgs = msgpack.unpackb(r)
                    for msg in msgs:
                        await self._dispatch(msg)
                except asyncio.TimeoutError:
                    # ws.recv is hanging when no data is received. by using
                    # wait_for we break when no data is received, allowing us
                    # to break the loop when needed
                    pass

    def _cast(self, msg_type, msg):
        result = msg
        if not self._raw_data:
            # convert msgpack timestamp to nanoseconds
            if 't' in msg:
                msg['t'] = msg['t'].seconds * int(1e9) + msg['t'].nanoseconds

            if msg_type == 't':
                result = Trade({
                    trade_mapping_v2[k]: v
                    for k, v in msg.items() if k in trade_mapping_v2
                })
            elif msg_type == 'q':
                result = Quote({
                    quote_mapping_v2[k]: v
                    for k, v in msg.items() if k in quote_mapping_v2
                })
            elif msg_type in ('b', 'u', 'd'):
                result = Bar({
                    bar_mapping_v2[k]: v
                    for k, v in msg.items() if k in bar_mapping_v2
                })
            else:
                result = Entity(msg)
        return result

    async def _dispatch(self, msg):
        msg_type = msg.get('T')
        symbol = msg.get('S')
        if msg_type == 't':
            handler = self._handlers['trades'].get(
                symbol, self._handlers['trades'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'q':
            handler = self._handlers['quotes'].get(
                symbol, self._handlers['quotes'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'b':
            handler = self._handlers['bars'].get(
                symbol, self._handlers['bars'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'u':
            handler = self._handlers['updatedBars'].get(
                symbol, self._handlers['updatedBars'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'd':
            handler = self._handlers['dailyBars'].get(
                symbol, self._handlers['dailyBars'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'subscription':
            sub = [f'{k}: {msg.get(k, [])}' for k in self._handlers]
            log.info(f'subscribed to {", ".join(sub)}')
        elif msg_type == 'error':
            log.error(f'error: {msg.get("msg")} ({msg.get("code")})')

    def _subscribe(self, handler, symbols, handlers):
        _ensure_coroutine(handler)
        for symbol in symbols:
            handlers[symbol] = handler
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._subscribe_all(), self._loop
            ).result()

    async def _subscribe_all(self):
        msg = defaultdict(list)
        for k, v in self._handlers.items():
            if k not in ("cancelErrors", "corrections") and v:
                for s in v.keys():
                    msg[k].append(s)
        msg['action'] = 'subscribe'
        bs = msgpack.packb(msg)
        frames = (bs[i:i+self._max_frame_size]
                  for i in range(0, len(bs), self._max_frame_size))
        await self._ws.send(frames)

    async def _unsubscribe(self,
                           trades=(),
                           quotes=(),
                           bars=(),
                           updated_bars=(),
                           daily_bars=()):
        raise NotImplementedError()

    async def _run_forever(self):
        self._loop = asyncio.get_running_loop()
        # do not start the websocket connection until we subscribe to something
        while not any(
            v for k, v in self._handlers.items()
            if k not in ("cancelErrors", "corrections")
        ):
            if not self._stop_stream_queue.empty():
                # the ws was signaled to stop before starting the loop so
                # we break
                self._stop_stream_queue.get(timeout=1)
                return
            await asyncio.sleep(0.1)
        log.info(f'started {self._name} stream')
        self._should_run = True
        self._running = False
        while True:
            try:
                if not self._should_run:
                    # when signaling to stop, this is how we break run_forever
                    log.info("{} stream stopped".format(self._name))
                    return
                if not self._running:
                    log.info("starting {} websocket connection".format(
                        self._name))
                    await self._start_ws()
                    await self._subscribe_all()
                    self._running = True
                await self._consume()
            except websockets.WebSocketException as wse:
                await self.close()
                self._running = False
                log.warn('data websocket error, restarting connection: ' +
                         str(wse))
            except Exception as e:
                log.exception('error during websocket '
                              'communication: {}'.format(str(e)))
            finally:
                await asyncio.sleep(0.01)

    def subscribe_trades(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['trades'])

    def subscribe_quotes(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['quotes'])

    def subscribe_bars(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['bars'])

    def subscribe_updated_bars(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['updatedBars'])

    def subscribe_daily_bars(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['dailyBars'])

    def unsubscribe_trades(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(trades=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['trades'][symbol]

    def unsubscribe_quotes(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(quotes=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['quotes'][symbol]

    def unsubscribe_bars(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(bars=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['bars'][symbol]

    def unsubscribe_updated_bars(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(updated_bars=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['updatedBars'][symbol]

    def unsubscribe_daily_bars(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(daily_bars=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['dailyBars'][symbol]

    def stop(self):
        if self._loop.is_running():
            asyncio.run_coroutine_threadsafe(
                self.stop_ws(),
                self._loop).result()


class DataStream(_DataStream):
    def __init__(self,
                 key_id: str,
                 secret_key: str,
                 base_url: URL,
                 raw_data: bool,
                 feed: str = 'iex',
                 websocket_params: Optional[Dict] = None):
        base_url = re.sub(r'^http', 'ws', base_url)
        super().__init__(endpoint=base_url + '/v2/' + feed,
                         key_id=key_id,
                         secret_key=secret_key,
                         raw_data=raw_data,
                         websocket_params=websocket_params
                         )
        self._handlers['statuses'] = {}
        self._handlers['lulds'] = {}
        self._handlers['cancelErrors'] = {}
        self._handlers['corrections'] = {}
        self._name = 'stock data'

    def _cast(self, msg_type, msg):
        result = super()._cast(msg_type, msg)
        if not self._raw_data:
            if msg_type == 's':
                result = StatusV2({
                    status_mapping_v2[k]: v
                    for k, v in msg.items() if k in status_mapping_v2
                })
            elif msg_type == 'l':
                result = LULDV2({
                    luld_mapping_v2[k]: v
                    for k, v in msg.items() if k in luld_mapping_v2
                })
            elif msg_type == 'x':
                result = CancelErrorV2({
                    cancel_error_mapping_v2[k]: v
                    for k, v in msg.items() if k in cancel_error_mapping_v2
                })
            elif msg_type == 'c':
                result = CorrectionV2({
                    correction_mapping_v2[k]: v
                    for k, v in msg.items() if k in correction_mapping_v2
                })
        return result

    async def _dispatch(self, msg):
        msg_type = msg.get('T')
        symbol = msg.get('S')
        if msg_type == 's':
            handler = self._handlers['statuses'].get(
                symbol, self._handlers['statuses'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'l':
            handler = self._handlers['lulds'].get(
                symbol, self._handlers['lulds'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'x':
            handler = self._handlers['cancelErrors'].get(
                symbol, self._handlers['cancelErrors'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        elif msg_type == 'c':
            handler = self._handlers['corrections'].get(
                symbol, self._handlers['corrections'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        else:
            await super()._dispatch(msg)

    async def _unsubscribe(self,
                           trades=(),
                           quotes=(),
                           bars=(),
                           updated_bars=(),
                           daily_bars=(),
                           statuses=(),
                           lulds=()):
        if (trades or quotes or bars or updated_bars or daily_bars or
                statuses or lulds):
            await self._ws.send(
                msgpack.packb({
                    'action':      'unsubscribe',
                    'trades':      trades,
                    'quotes':      quotes,
                    'bars':        bars,
                    'updatedBars': updated_bars,
                    'dailyBars':   daily_bars,
                    'statuses':    statuses,
                    'lulds':       lulds,
                }))

    def subscribe_statuses(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['statuses'])

    def subscribe_lulds(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['lulds'])

    def unsubscribe_statuses(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(statuses=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['statuses'][symbol]

    def unsubscribe_lulds(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(lulds=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['lulds'][symbol]

    def register_handler(self, msg_type, handler, *symbols):
        if handler is not None:
            _ensure_coroutine(handler)
            for symbol in symbols:
                self._handlers[msg_type][symbol] = handler

    def unregister_handler(self, msg_type, *symbols):
        for symbol in symbols:
            if symbol in self._handlers[msg_type]:
                del self._handlers[msg_type][symbol]


class CryptoDataStream(_DataStream):
    def __init__(self,
                 key_id: str,
                 secret_key: str,
                 base_url: URL,
                 raw_data: bool,
                 exchanges: Optional[List[str]] = None,
                 websocket_params: Optional[Dict] = None):
        self._key_id = key_id
        self._secret_key = secret_key
        base_url = re.sub(r'^http', 'ws', base_url)
        endpoint = base_url + '/v1beta2/crypto'
        if exchanges:
            if isinstance(exchanges, str):
                endpoint += '?exchanges=' + exchanges
            else:
                endpoint += '?exchanges=' + ','.join(exchanges)
        super().__init__(endpoint=endpoint,
                         key_id=key_id,
                         secret_key=secret_key,
                         raw_data=raw_data,
                         websocket_params=websocket_params,
                         )
        self._handlers['orderbooks'] = {}
        self._name = 'crypto data'

    def _cast(self, msg_type, msg):
        result = super()._cast(msg_type, msg)
        if not self._raw_data:
            if msg_type == 'o':
                result = OrderbookV2({
                    orderbook_mapping_v2[k]: v
                    for k, v in msg.items() if k in orderbook_mapping_v2
                })
        return result

    async def _dispatch(self, msg):
        msg_type = msg.get('T')
        symbol = msg.get('S')
        if msg_type == 'o':
            handler = self._handlers['orderbooks'].get(
                symbol, self._handlers['orderbooks'].get('*', None))
            if handler:
                await handler(self._cast(msg_type, msg))
        else:
            await super()._dispatch(msg)

    async def _unsubscribe(self,
                           trades=(),
                           quotes=(),
                           orderbooks=(),
                           bars=(),
                           updated_bars=(),
                           daily_bars=()):
        if (
            trades or quotes or orderbooks or bars or updated_bars
            or daily_bars
        ):
            await self._ws.send(
                msgpack.packb({
                    'action':      'unsubscribe',
                    'trades':      trades,
                    'quotes':      quotes,
                    'orderbooks':  orderbooks,
                    'bars':        bars,
                    'updatedBars': updated_bars,
                    'dailyBars':   daily_bars,
                }))

    def subscribe_orderbooks(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['orderbooks'])

    def unsubscribe_orderbooks(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(orderbooks=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['orderbooks'][symbol]


class NewsDataStream(_DataStream):
    def __init__(self,
                 key_id: str,
                 secret_key: str,
                 base_url: URL,
                 raw_data: bool,
                 websocket_params: Optional[Dict] = None):
        self._key_id = key_id
        self._secret_key = secret_key
        base_url = re.sub(r'^http', 'ws', base_url)
        endpoint = base_url + '/v1beta1/news'
        super().__init__(endpoint=endpoint,
                         key_id=key_id,
                         secret_key=secret_key,
                         raw_data=raw_data,
                         websocket_params=websocket_params
                         )
        self._handlers = {
            'news':    {},
        }
        self._name = 'news data'

    def _cast(self, msg_type, msg):
        result = super()._cast(msg_type, msg)
        if not self._raw_data:
            if msg_type == 'n':
                result = NewsV2(msg)
        return result

    async def _dispatch(self, msg):
        msg_type = msg.get('T')
        if msg_type == 'n':
            symbols = msg.get('symbols', [])
            # A news article could be unrelated to any symbols,
            # resulting in an empty symbols list. Those news articles
            # should still be dispatched to the wildcard event handler.
            if not symbols:
                symbols.append('*')

            handlers = set()
            for symbol in symbols:
                handler = self._handlers['news'].get(symbol)
                if handler is None:
                    handler = self._handlers['news'].get('*')
                if handler is not None:
                    handlers.add(handler)
            for handler in handlers:
                await handler(self._cast(msg_type, msg))
        else:
            await super()._dispatch(msg)

    async def _unsubscribe(self, news=()):
        if news:
            await self._ws.send(
                msgpack.packb({
                    'action': 'unsubscribe',
                    'news':    news,
                }))

    def subscribe_news(self, handler, *symbols):
        self._subscribe(handler, symbols, self._handlers['news'])

    def unsubscribe_news(self, *symbols):
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._unsubscribe(news=symbols),
                self._loop).result()
        for symbol in symbols:
            del self._handlers['news'][symbol]


class TradingStream:
    def __init__(self,
                 key_id: str,
                 secret_key: str,
                 base_url: URL,
                 raw_data: bool = False,
                 websocket_params: Optional[Dict] = None):
        self._key_id = key_id
        self._secret_key = secret_key
        base_url = re.sub(r'^http', 'ws', base_url)
        self._endpoint = base_url + '/stream/'
        self._trade_updates_handler = None
        self._ws = None
        self._running = False
        self._loop = None
        self._raw_data = raw_data
        self._stop_stream_queue = queue.Queue()
        self._should_run = True
        self._websocket_params = websocket_params

        if self._websocket_params is None:
            self._websocket_params = WEBSOCKET_DEFAULTS

    async def _connect(self):
        self._ws = await websockets.connect(
            self._endpoint,
            **self._websocket_params
        )

    async def _auth(self):
        await self._ws.send(
            json.dumps({
                'action': 'authenticate',
                'data':   {
                    'key_id':     self._key_id,
                    'secret_key': self._secret_key,
                }
            }))
        r = await self._ws.recv()
        msg = json.loads(r)
        if msg.get('data').get('status') != 'authorized':
            raise ValueError('failed to authenticate')

    async def _dispatch(self, msg):
        stream = msg.get('stream')
        if stream == 'trade_updates':
            if self._trade_updates_handler:
                await self._trade_updates_handler(self._cast(msg))

    def _cast(self, msg):
        result = msg
        if not self._raw_data:
            result = Entity(msg.get('data'))
        return result

    async def _subscribe_trade_updates(self):
        if self._trade_updates_handler:
            await self._ws.send(
                json.dumps({
                    'action': 'listen',
                    'data':   {
                        'streams': ['trade_updates']
                    }
                }))

    def subscribe_trade_updates(self, handler):
        _ensure_coroutine(handler)
        self._trade_updates_handler = handler
        if self._running:
            asyncio.run_coroutine_threadsafe(
                self._subscribe_trade_updates(),
                self._loop).result()

    async def _start_ws(self):
        await self._connect()
        await self._auth()
        log.info(f'connected to: {self._endpoint}')
        await self._subscribe_trade_updates()

    async def _consume(self):
        while True:
            if not self._stop_stream_queue.empty():
                self._stop_stream_queue.get(timeout=1)
                await self.close()
                break
            else:
                try:
                    r = await asyncio.wait_for(self._ws.recv(), 5)
                    msg = json.loads(r)
                    await self._dispatch(msg)
                except asyncio.TimeoutError:
                    # ws.recv is hanging when no data is received. by using
                    # wait_for we break when no data is received, allowing us
                    # to break the loop when needed
                    pass

    async def _run_forever(self):
        self._loop = asyncio.get_running_loop()
        # do not start the websocket connection until we subscribe to something
        while not self._trade_updates_handler:
            if not self._stop_stream_queue.empty():
                self._stop_stream_queue.get(timeout=1)
                return
            await asyncio.sleep(0.1)
        log.info('started trading stream')
        self._should_run = True
        self._running = False
        while True:
            try:
                if not self._should_run:
                    log.info("Trading stream stopped")
                    return
                if not self._running:
                    log.info("starting trading websocket connection")
                    await self._start_ws()
                    self._running = True
                    await self._consume()
            except websockets.WebSocketException as wse:
                await self.close()
                self._running = False
                log.warn('trading stream websocket error, restarting ' +
                         ' connection: ' + str(wse))
            except Exception as e:
                log.exception('error during websocket '
                              'communication: {}'.format(str(e)))
            finally:
                await asyncio.sleep(0.01)

    async def close(self):
        if self._ws:
            await self._ws.close()
            self._ws = None
            self._running = False

    async def stop_ws(self):
        self._should_run = False
        if self._stop_stream_queue.empty():
            self._stop_stream_queue.put_nowait({"should_stop": True})

    def stop(self):
        if self._loop.is_running():
            asyncio.run_coroutine_threadsafe(
                self.stop_ws(),
                self._loop).result()


class Stream:
    def __init__(self,
                 key_id: str = None,
                 secret_key: str = None,
                 base_url: URL = None,
                 data_stream_url: URL = None,
                 data_feed: str = 'iex',
                 raw_data: bool = False,
                 crypto_exchanges: Optional[List[str]] = None,
                 websocket_params: Optional[Dict] = None):
        self._key_id, self._secret_key, _ = get_credentials(key_id, secret_key)
        self._base_url = base_url or get_base_url()
        self._data_stream_url = data_stream_url or get_data_stream_url()

        self._trading_ws = TradingStream(self._key_id,
                                         self._secret_key,
                                         self._base_url,
                                         raw_data,
                                         websocket_params=websocket_params)
        self._data_ws = DataStream(self._key_id,
                                   self._secret_key,
                                   self._data_stream_url,
                                   raw_data,
                                   data_feed.lower(),
                                   websocket_params=websocket_params)
        self._crypto_ws = CryptoDataStream(self._key_id,
                                           self._secret_key,
                                           self._data_stream_url,
                                           raw_data,
                                           crypto_exchanges,
                                           websocket_params=websocket_params)
        self._news_ws = NewsDataStream(self._key_id,
                                       self._secret_key,
                                       self._data_stream_url,
                                       raw_data,
                                       websocket_params=websocket_params)

    def subscribe_trade_updates(self, handler):
        self._trading_ws.subscribe_trade_updates(handler)

    def subscribe_trades(
        self,
        handler,
        *symbols,
        handler_cancel_errors=None,
        handler_corrections=None
    ):
        self._data_ws.subscribe_trades(handler, *symbols)
        self._data_ws.register_handler("cancelErrors",
                                       handler_cancel_errors,
                                       *symbols)
        self._data_ws.register_handler("corrections",
                                       handler_corrections,
                                       *symbols)

    def subscribe_quotes(self, handler, *symbols):
        self._data_ws.subscribe_quotes(handler, *symbols)

    def subscribe_bars(self, handler, *symbols):
        self._data_ws.subscribe_bars(handler, *symbols)

    def subscribe_updated_bars(self, handler, *symbols):
        self._data_ws.subscribe_updated_bars(handler, *symbols)

    def subscribe_daily_bars(self, handler, *symbols):
        self._data_ws.subscribe_daily_bars(handler, *symbols)

    def subscribe_statuses(self, handler, *symbols):
        self._data_ws.subscribe_statuses(handler, *symbols)

    def subscribe_lulds(self, handler, *symbols):
        self._data_ws.subscribe_lulds(handler, *symbols)

    def subscribe_crypto_trades(self, handler, *symbols):
        self._crypto_ws.subscribe_trades(handler, *symbols)

    def subscribe_crypto_quotes(self, handler, *symbols):
        self._crypto_ws.subscribe_quotes(handler, *symbols)

    def subscribe_crypto_bars(self, handler, *symbols):
        self._crypto_ws.subscribe_bars(handler, *symbols)

    def subscribe_crypto_updated_bars(self, handler, *symbols):
        self._crypto_ws.subscribe_updated_bars(handler, *symbols)

    def subscribe_crypto_daily_bars(self, handler, *symbols):
        self._crypto_ws.subscribe_daily_bars(handler, *symbols)

    def subscribe_crypto_orderbooks(self, handler, *symbols):
        self._crypto_ws.subscribe_orderbooks(handler, *symbols)

    def subscribe_news(self, handler, *symbols):
        self._news_ws.subscribe_news(handler, *symbols)

    def on_trade_update(self, func):
        self.subscribe_trade_updates(func)
        return func

    def on_trade(self, *symbols):
        def decorator(func):
            self.subscribe_trades(func, *symbols)
            return func

        return decorator

    def on_quote(self, *symbols):
        def decorator(func):
            self.subscribe_quotes(func, *symbols)
            return func

        return decorator

    def on_bar(self, *symbols):
        def decorator(func):
            self.subscribe_bars(func, *symbols)
            return func

        return decorator

    def on_updated_bar(self, *symbols):
        def decorator(func):
            self.subscribe_updated_bars(func, *symbols)
            return func

        return decorator

    def on_daily_bar(self, *symbols):
        def decorator(func):
            self.subscribe_daily_bars(func, *symbols)
            return func

        return decorator

    def on_status(self, *symbols):
        def decorator(func):
            self.subscribe_statuses(func, *symbols)
            return func

        return decorator

    def on_luld(self, *symbols):
        def decorator(func):
            self.subscribe_lulds(func, *symbols)
            return func

        return decorator

    def on_cancel_error(self, *symbols):
        def decorator(func):
            self._data_ws.register_handler("cancelErrors", func, *symbols)
            return func

        return decorator

    def on_correction(self, *symbols):
        def decorator(func):
            self._data_ws.register_handler("corrections", func, *symbols)
            return func

        return decorator

    def on_crypto_trade(self, *symbols):
        def decorator(func):
            self.subscribe_crypto_trades(func, *symbols)
            return func

        return decorator

    def on_crypto_quote(self, *symbols):
        def decorator(func):
            self.subscribe_crypto_quotes(func, *symbols)
            return func

        return decorator

    def on_crypto_bar(self, *symbols):
        def decorator(func):
            self.subscribe_crypto_bars(func, *symbols)
            return func

        return decorator

    def on_crypto_updated_bar(self, *symbols):
        def decorator(func):
            self.subscribe_crypto_updated_bars(func, *symbols)
            return func

        return decorator

    def on_crypto_daily_bar(self, *symbols):
        def decorator(func):
            self.subscribe_crypto_daily_bars(func, *symbols)
            return func

        return decorator

    def on_crypto_orderbook(self, *symbols):
        def decorator(func):
            self.subscribe_crypto_orderbooks(func, *symbols)
            return func

        return decorator

    def on_news(self, *symbols):
        def decorator(func):
            self.subscribe_news(func, *symbols)
            return func

        return decorator

    def unsubscribe_trades(self, *symbols):
        self._data_ws.unsubscribe_trades(*symbols)
        self._data_ws.unregister_handler("cancelErrors", *symbols)
        self._data_ws.unregister_handler("corrections", *symbols)

    def unsubscribe_quotes(self, *symbols):
        self._data_ws.unsubscribe_quotes(*symbols)

    def unsubscribe_bars(self, *symbols):
        self._data_ws.unsubscribe_bars(*symbols)

    def unsubscribe_updated_bars(self, *symbols):
        self._data_ws.unsubscribe_updated_bars(*symbols)

    def unsubscribe_daily_bars(self, *symbols):
        self._data_ws.unsubscribe_daily_bars(*symbols)

    def unsubscribe_statuses(self, *symbols):
        self._data_ws.unsubscribe_statuses(*symbols)

    def unsubscribe_lulds(self, *symbols):
        self._data_ws.unsubscribe_lulds(*symbols)

    def unsubscribe_crypto_trades(self, *symbols):
        self._crypto_ws.unsubscribe_trades(*symbols)

    def unsubscribe_crypto_quotes(self, *symbols):
        self._crypto_ws.unsubscribe_quotes(*symbols)

    def unsubscribe_crypto_bars(self, *symbols):
        self._crypto_ws.unsubscribe_bars(*symbols)

    def unsubscribe_crypto_updated_bars(self, *symbols):
        self._crypto_ws.unsubscribe_updated_bars(*symbols)

    def unsubscribe_crypto_daily_bars(self, *symbols):
        self._crypto_ws.unsubscribe_daily_bars(*symbols)

    def unsubscribe_crypto_orderbooks(self, *symbols):
        self._crypto_ws.unsubscribe_orderbooks(*symbols)

    def unsubscribe_news(self, *symbols):
        self._news_ws.unsubscribe_news(*symbols)

    async def _run_forever(self):
        await asyncio.gather(self._trading_ws._run_forever(),
                             self._data_ws._run_forever(),
                             self._crypto_ws._run_forever(),
                             self._news_ws._run_forever())

    def run(self):
        try:
            asyncio.run(self._run_forever())
        except KeyboardInterrupt:
            print('keyboard interrupt, bye')
            pass

    async def stop_ws(self):
        """
        Signal the ws connections to stop listenning to api stream.
        """
        if self._trading_ws:
            await self._trading_ws.stop_ws()

        if self._data_ws:
            await self._data_ws.stop_ws()

        if self._crypto_ws:
            await self._crypto_ws.stop_ws()

        if self._news_ws:
            await self._news_ws.stop_ws()

    def stop(self):
        if self._trading_ws:
            self._trading_ws.stop()
        if self._data_ws:
            self._data_ws.stop()
        if self._crypto_ws:
            self._crypto_ws.stop()
        if self._news_ws:
            self._news_ws.stop()

    def is_open(self):
        """
        Checks if either of the websockets is open
        :return:
        """
        open_ws = (self._trading_ws._ws or self._data_ws._ws
                   or self._crypto_ws._ws or self._news_ws)  # noqa
        if open_ws:
            return True
        return False



from enum import Enum
import pandas as pd
from .entity import Bar, Entity, Trade, Quote, _NanoTimestamped
from typing import Dict

trade_mapping_v2 = {
    "i": "id",
    "S": "symbol",
    "c": "conditions",
    "x": "exchange",
    "p": "price",
    "s": "size",
    "t": "timestamp",
    "z": "tape",  # stocks only
    "tks": "takerside"  # crypto only
}

quote_mapping_v2 = {
    "S":  "symbol",
    "x": "exchange",  # crypto only
    "ax": "ask_exchange",
    "ap": "ask_price",
    "as": "ask_size",
    "bx": "bid_exchange",
    "bp": "bid_price",
    "bs": "bid_size",
    "c":  "conditions",  # stocks only
    "t":  "timestamp",
    "z":  "tape"  # stocks only
}

bar_mapping_v2 = {
    "S":  "symbol",
    "x": "exchange",  # crypto only
    "o":  "open",
    "h":  "high",
    "l":  "low",
    "c":  "close",
    "v":  "volume",
    "t":  "timestamp",
    "n":  "trade_count",
    "vw": "vwap"
}

status_mapping_v2 = {
    "S":  "symbol",
    "sc": "status_code",
    "sm": "status_message",
    "rc": "reason_code",
    "rm": "reason_message",
    "t":  "timestamp",
    "z":  "tape"
}

luld_mapping_v2 = {
    "S": "symbol",
    "u": "limit_up_price",
    "d": "limit_down_price",
    "i": "indicator",
    "t": "timestamp",
    "z": "tape"
}

cancel_error_mapping_v2 = {
    "S": "symbol",
    "i": "id",
    "x": "exchange",
    "p": "price",
    "s": "size",
    "a": "cancel_error_action",
    "z": "tape",
    "t": "timestamp",
}

correction_mapping_v2 = {
    "S": "symbol",
    "x": "exchange",
    "oi": "original_id",
    "op": "original_price",
    "os": "original_size",
    "oc": "original_conditions",
    "ci": "corrected_id",
    "cp": "corrected_price",
    "cs": "corrected_size",
    "cc": "corrected_conditions",
    "z": "tape",
    "t": "timestamp",
}

orderbook_mapping_v2 = {
    "S": "symbol",
    "x": "exchange",
    "t": "timestamp",
    "b": "bids",
    "a": "asks",
}


class EntityListType(Enum):
    Trade = Trade, trade_mapping_v2
    Quote = Quote, quote_mapping_v2
    Bar = Bar, bar_mapping_v2


class EntityList(list):
    def __init__(self, entity_type: EntityListType, raw):
        entity = entity_type.value[0]
        super().__init__([entity(o) for o in raw])
        self._raw = raw
        self.mapping = entity_type.value[1]

    @property
    def df(self):
        if not hasattr(self, '_df'):
            df = pd.DataFrame(
                self._raw,
            )

            df.columns = [self.mapping.get(c, c) for c in df.columns]
            if not df.empty:
                df.set_index('timestamp', inplace=True)
                df.index = pd.DatetimeIndex(df.index)
            self._df = df
        return self._df


class Remapped:
    def __init__(self, mapping: Dict[str, str], *args, **kwargs):
        self._reversed_mapping = {
            value: key for (key, value) in mapping.items()}
        super().__init__(*args, **kwargs)

    def __getattr__(self, key):
        if key in self._reversed_mapping:
            return super().__getattr__(self._reversed_mapping[key])
        return super().__getattr__(key)


class BarsV2(EntityList):
    def __init__(self, raw):
        super().__init__(EntityListType.Bar, raw)


class TradesV2(EntityList):
    def __init__(self, raw):
        super().__init__(EntityListType.Trade, raw)


class QuotesV2(EntityList):
    def __init__(self, raw):
        super().__init__(EntityListType.Quote, raw)


class TradeV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(trade_mapping_v2, raw)


class QuoteV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(quote_mapping_v2, raw)


class BarV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(bar_mapping_v2, raw)


class StatusV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(status_mapping_v2, raw)


class LULDV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(luld_mapping_v2, raw)


class CancelErrorV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(cancel_error_mapping_v2, raw)


class CorrectionV2(Remapped, _NanoTimestamped, Entity):
    _tskeys = ('t',)

    def __init__(self, raw):
        super().__init__(correction_mapping_v2, raw)


class SnapshotV2:
    def __init__(self, raw):
        self.latest_trade = _convert_or_none(TradeV2, raw.get('latestTrade'))
        self.latest_quote = _convert_or_none(QuoteV2, raw.get('latestQuote'))
        self.minute_bar = _convert_or_none(BarV2, raw.get('minuteBar'))
        self.daily_bar = _convert_or_none(BarV2, raw.get('dailyBar'))
        self.prev_daily_bar = _convert_or_none(BarV2, raw.get('prevDailyBar'))


class SnapshotsV2(dict):
    def __init__(self, raw):
        for k, v in raw.items():
            self[k] = _convert_or_none(SnapshotV2, v)


class LatestBarsV2(dict):
    def __init__(self, raw):
        for k, v in raw.items():
            self[k] = _convert_or_none(BarV2, v)


class LatestTradesV2(dict):
    def __init__(self, raw):
        for k, v in raw.items():
            self[k] = _convert_or_none(TradeV2, v)


class LatestQuotesV2(dict):
    def __init__(self, raw):
        for k, v in raw.items():
            self[k] = _convert_or_none(QuoteV2, v)


class BidOrAsk(Entity):
    def __init__(self, raw):
        super().__init__(raw)


class OrderbookV2(Entity):
    def __init__(self, raw):
        super().__init__(raw)
        if self.bids:
            for i in range(len(self.bids)):
                self.bids[i] = BidOrAsk(self.bids[i])
        if self.asks:
            for i in range(len(self.asks)):
                self.asks[i] = BidOrAsk(self.asks[i])


class OrderbooksV2(dict):
    def __init__(self, raw):
        for k, v in raw.items():
            for side in orderbook_mapping_v2.keys():
                if side not in v:
                    continue
                readable_side = orderbook_mapping_v2[side]
                v[readable_side] = v[side]
                v.pop(side)
            self[k] = _convert_or_none(OrderbookV2, v)


class NewsV2(Entity):
    def __init__(self, raw):
        super().__init__(raw)


class NewsListV2(list):
    def __init__(self, raw):
        super().__init__([NewsV2(o) for o in raw])


def _convert_or_none(entityType, value):
    if value:
        return entityType(value)
    return None
